# coding=utf-8
from __future__ import unicode_literals

import logging

import related


class CharsetConverter:
    def _upper_replacements(self, replacements):
        upper_replacements = []
        for replacement in replacements:
            try:
                c_from, c_to = replacement["from"], replacement["to"]
                upper_replacements.append({
                    "from": c_from.upper(),
                    "to": c_to.upper()
                })
            except (KeyError, ValueError, TypeError):
                self.log.exception("skipping incorrect configured replacement: {}".format(replacement))
                continue
        return upper_replacements

    def __init__(self, replacements, allowed_chars):
        self.log = logging.getLogger("charset")
        self.replacements = self._upper_replacements(replacements)
        self.allowed_chars = allowed_chars.upper()

    @classmethod
    def from_yaml(cls, charset):
        allowed_chars = charset.allowed
        replacements = [{"from": r.c_from, "to": r.c_to} for r in charset.replacements]
        return cls(replacements, allowed_chars)

    def _replace(self, text):
        for replacement in self.replacements:
            c_from, c_to = replacement.get("from"), replacement.get("to")
            text = text.replace(c_from, c_to)
        return text

    def _remove_unknown(self, text):
        return "".join(x for x in text if x in self.allowed_chars)

    def clean(self, text):
        return self._remove_unknown(self._replace(text.upper()))


@related.immutable
class Replacement(object):
    c_from = related.StringField(key="from", required=True)
    c_to = related.StringField(key="to", required=True)


default_replacements = [
    Replacement(c_from=c_from, c_to=c_to) for (c_from, c_to) in
    (("Ä", "A"), ("Ü", "U"), ("Ö", "O"), ("O", "0"), ("Q", "0"))
]


def CharsetField():
    return related.ChildField(Charset, required=False, default=Charset())


@related.immutable
class Charset(object):
    replacements = related.SequenceField(Replacement, default=default_replacements)
    allowed = related.StringField(default="0123456789abcdefghijklmnprstuvwxyz")


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    charset = Charset(replacements=[{"from": 'ä', "to": 'a'}], allowed_chars="Ac")
    print(charset.clean("ÄAÖBCÄDD"))
    assert charset.clean("ÄAÖBCÄDD") == "AACA"



# coding=utf-8
from __future__ import unicode_literals

import os
import sys
import threading
import time
import traceback

import zmq

from om.message import Topic, Message


class HealthzStatus:
    OK = "ok"
    ERROR = "error"


startup = time.time()


class Healthz(object):
    def __init__(self, config=None, socket=None):
        self.pub_socket = None
        self.version = "unknown"
        self.name = "unknown"
        self.socket = socket

        if socket:
            self.subscribe(socket)

        if config:
            self.connect(config.BROKER_HOST, config.BROKER_SUB_PORT)
            self.version = config.VERSION
            self.name = config.NAME

    def connect(self, broker_host, broker_sub_port):
        context = zmq.Context()
        self.pub_socket = context.socket(zmq.PUB)
        self.pub_socket.connect("tcp://{}:{}".format(broker_host, broker_sub_port))

    def status_dict(self, status, _message=None, **meta):
        result = {
            "status": status,
            "meta": meta,
            "version": self.version,
            "name": self.name,
            "startup": startup
        }
        if _message:
            result["message"] = _message
        return result

    def ok(self, **meta):
        return self.status_dict(HealthzStatus.OK, **meta)

    def error(self, message, **meta):
        return self.status_dict(HealthzStatus.ERROR, message, **meta)

    def subscribe(self, socket):
        socket.subscribe(Topic.HEALTH)

    def process(self, topic, message):
        if topic != Topic.HEALTH:
            return False
        if "ping" not in message:
            return True
        message = Message(
            name=self.name,
            type="healthz",
            pong=self.check()
        )
        message.to_socket(self.pub_socket, Topic.HEALTH)
        return True

    def check(self):
        return self.ok()

    def _run(self):
        try:
            time.sleep(0.5)
            while True:
                topic, message = Message.from_socket(self.socket)
                self.process(topic, message)
        except:
            traceback.print_exc()
        finally:
            sys.stderr.write("error in health thread")
            sys.stderr.flush()
            os._exit(4)

    @classmethod
    def as_thread(cls, config):
        context = zmq.Context()
        socket = context.socket(zmq.SUB)
        socket.connect("tcp://{}:{}".format(config.BROKER_HOST, config.BROKER_PUB_PORT))

        health = cls(config=config, socket=socket)
        thread = threading.Thread(target=health._run)
        thread.setDaemon(True)

        return thread


class RedisHealthz(Healthz):
    def __init__(self, redis, config=None, socket=None):
        super(RedisHealthz, self).__init__(config=config, socket=socket)
        self.redis = redis

    def check(self):
        try:
            ok = self.redis.ping()
        except Exception as e:
            return self.error("redis server not reachable: {}".format(str(e)))

        return self.ok() if ok else self.error("redis server not reachable")



# coding=utf-8
from __future__ import unicode_literals

import glob
import logging
import os
import sys

import editdistance
import yaml

SEPARATOR_CHARS = " -_.:"

if sys.version_info < (3, 0):
    def ucode(txt):
        return unicode(txt)
else:
    def ucode(txt):
        return str(txt)


class PlateYaml(object):
    def __init__(self, filename):
        self.scheme = ""
        self.version = -1
        self.exactitude = 0.0
        self.filename = filename
        with open(filename, 'r') as f:
            try:
                self.settings = yaml.load(f)
                self.valid_chars = self.settings["allowed"]
                self.exactitude = self.settings["exactitude"]
                self.kept_separators = self.settings.get("kept_separators", "NONE").upper()
                self.replacements = []
                for r in self.settings.get("replacements", []):
                    if r.get("from", "") != "" and "to" in r:
                        self.replacements.append((ucode(r.get("from")), ucode(r.get("to"))))
            except:
                logging.error("Error loading Plate Yaml %s, will be ignored", filename)
                return

        try:
            self.scheme, tmp = os.path.basename(filename).replace(".", "-").split("-")[:2]
            self.version = int(tmp)
        except Exception as e:
            logging.exception(e)
            logging.error("Invalid Plate Yaml filename %s, will be ignored", filename)
            return

    def convert(self, plate):
        plate.upper()
        for r in self.replacements:
            plate = plate.replace(r[0], r[1])
        tmp = plate
        for c in tmp:
            if c not in self.valid_chars:
                plate = plate.replace(c, "")

        if self.kept_separators != "ALL":
            plate_stripped = plate
            for c in list(SEPARATOR_CHARS):
                plate_stripped = plate_stripped.replace(c, "")
            if self.kept_separators == "FIRST":
                for i, c in enumerate(list(plate)):
                    if c in SEPARATOR_CHARS:
                        plate = plate[:i + 1] + plate_stripped[i:]
                        break
            elif self.kept_separators == "LAST":
                for i, c in enumerate(reversed(list(plate))):
                    if c in SEPARATOR_CHARS:
                        plate = plate_stripped[:-i] + plate[-1 - i:]
                        break
            else:
                plate = plate_stripped

        return plate


class PlateYamlSelector(object):
    def __init__(self, yaml_dir, default_country_order=[], legacy_in_default_country_order=False):
        self.yaml_dir = yaml_dir
        self.default_country_order = default_country_order
        self.legacy_in_default_country_order = legacy_in_default_country_order
        yaml_filenames = glob.glob(os.path.join(yaml_dir, "*.yml"))
        self.schemes = {}
        for fn in yaml_filenames:
            plate_yaml = PlateYaml(fn)
            if plate_yaml.version >= 0:  # check, if loading was successful
                if plate_yaml.scheme in self.schemes:
                    self.schemes[plate_yaml.scheme][plate_yaml.version] = plate_yaml
                else:
                    self.schemes[plate_yaml.scheme] = {plate_yaml.version: plate_yaml}
        if self.schemes.get("LEGACY", {}).get(0, None) is None:
            raise Exception("NO LEGACY-0.yml")

        legacy_yaml = self.schemes["LEGACY"][0]
        del self.schemes["LEGACY"]
        if "DEFAULT" not in self.schemes:
            self.schemes["DEFAULT"] = {}

        for key in self.schemes.keys():
            self.schemes[key][0] = legacy_yaml

    def get_scheme_versions_decreasing(self, country):
        if country not in self.schemes:
            country = "DEFAULT"
        return sorted(self.schemes[country].keys(), key=lambda x: self.schemes[country][x].exactitude, reverse=True)

    def get_all_exactitudes_decreasing(self):
        exactitudes = set([])
        for s in self.schemes.keys():
            for v in self.schemes[s].keys():
                exactitudes.add(self.schemes[s][v].exactitude)
        return sorted(exactitudes, reverse=True)

    def get_best_yaml(self, scheme, version):
        if scheme not in self.schemes:
            scheme = "DEFAULT"

        versions = sorted(self.schemes.get(scheme, {}).keys(), reverse=True)
        for v in versions:
            if v <= version:
                return self.schemes[scheme][v]

        raise Exception("NO YAML FOUND! THIS SHOULD NOT HAPPEN!")

    def convert(self, plate, scheme, version, return_matched_yaml=False):
        if return_matched_yaml:
            best_yaml = self.get_best_yaml(scheme, version)
            return {"plate": best_yaml.convert(plate), "scheme": best_yaml.scheme, "version": best_yaml.version,
                    "exactitude": best_yaml.exactitude}
        else:
            return self.get_best_yaml(scheme, version).convert(plate)

    def check_ed0(self, db_cv, plate, matches, num_alt, return_only_best=True):
        if db_cv and "plate_scheme" in db_cv[0]:
            p_conv = self.convert(plate, db_cv[0]["used_scheme"], db_cv[0]["used_version"])
            for db_entry in db_cv:
                if p_conv == db_entry["plate"]:
                    matches[num_alt].append(db_entry)
                    if return_only_best:
                        return db_entry, matches
        return None, matches

    def check_edx(self, db_cv, plate, max_ed, matches, return_only_best=True):
        if db_cv and "plate_scheme" in db_cv[0]:
            p_conv = self.convert(plate, db_cv[0]["used_scheme"], db_cv[0]["used_version"])
            for db_entry in db_cv:
                ed = editdistance.eval(p_conv, db_entry["plate"])
                if ed <= max_ed:
                    matches[ed].append(db_entry)
                if ed == 0 and return_only_best:
                    return db_entry, matches
        return None, matches

    def match_plate(self, plate, db, use_alternatives=True, max_ed=0, return_only_best=True):
        """
        Expected format of plate: {"plate": {"plate": <plate>, ...},
                                       "country": [{"country": <A/D/...>, "confidence": <0.0-1.0>}],
                                       "alternatives": [{"plate": <alt_plate1>, "confidence": <0.0-1.0>},...]}
        Expected format of db: {<Country1>: {<Version1>: [<plate1>, <plate2>,...]}}
            plates in converted form (converted using the best matching yaml, sorted into convert scheme and version,
                                                                                not db entry scheme and version)
        ED0 if use_alternatives (ed1 and alternatives does not make much sense)
        """
        if use_alternatives:
            max_ed = 0
            plates = plate.get("alternatives", [])
            matches = [[]] * len(plates)
        else:
            plates = plate.get("alternatives", [])[:1]
            matches = [[]] * (max_ed + 1)
        for num_alt, p in enumerate(plates):
            check_other = False
            # first check based on countries given in plate
            checked_countries = []
            for c in plate.get("country", []):
                country = c["country"]
                if country == "OTHER":
                    check_other = True
                    break
                if country not in db:
                    continue
                for version in self.get_scheme_versions_decreasing(country):
                    if max_ed == 0:
                        res, matches = self.check_ed0(db[country].get(version, []), p["plate"], matches, num_alt,
                                                      return_only_best)
                    else:
                        res, matches = self.check_edx(db[country].get(version, []), p["plate"], max_ed, matches,
                                                      return_only_best)
                    if res is not None:
                        return [[res]]
                checked_countries.append(country)

            # if other specified, check other contries
            if check_other:
                no_legacy_checked_countries = []
                # first check in order given in config, but ignoring legacy (if not legacy_in_default_country_order)
                for country in self.default_country_order:
                    if country in checked_countries or country not in db:
                        continue
                    versions = self.get_scheme_versions_decreasing(country)
                    if not self.legacy_in_default_country_order:
                        versions = versions[:-1]
                    for version in versions:
                        if max_ed == 0:
                            res, matches = self.check_ed0(db[country].get(version, []), p["plate"], matches, num_alt,
                                                          return_only_best)
                        else:
                            res, matches = self.check_edx(db[country].get(version, []), p["plate"], max_ed, matches,
                                                          return_only_best)
                        if res is not None:
                            return [[res]]
                    if self.legacy_in_default_country_order:
                        checked_countries.append(country)
                    else:
                        no_legacy_checked_countries.append(country)

                # rest is checked in decreasing version order
                for exactitude in self.get_all_exactitudes_decreasing():
                    for country, db_country in db.items():
                        if country in checked_countries or exactitude != 0 and country in no_legacy_checked_countries:
                            continue
                        for version, db_cv in db_country.items():
                            if db_cv and db_cv[0]["exactitude"] == exactitude:
                                if max_ed == 0:
                                    res, matches = self.check_ed0(db[country].get(version, []), p["plate"], matches,
                                                                  num_alt, return_only_best)
                                else:
                                    res, matches = self.check_edx(db[country].get(version, []), p["plate"], max_ed,
                                                                  matches, return_only_best)
                                if res is not None:
                                    return [[res]]

            if return_only_best:
                for edm in matches:
                    if len(edm) > 0:
                        return [[edm[0]]]

        return matches



# coding=utf-8
from __future__ import unicode_literals

# builtins is python 3.x feature
import __builtin__
import logging
import os

import related
import six


def _save_cast(obj, key, to_type, default):
    try:
        return to_type(obj[key])
    except (ValueError, TypeError, KeyError):
        return default


def int(key, default=0):
    assert isinstance(default, (__builtin__.int, __builtin__.float)), "default argument must be of type int or float"
    return _save_cast(os.environ, key, __builtin__.int, __builtin__.int(default))


def float(key, default=0.0):
    assert isinstance(default, (__builtin__.int, __builtin__.float)), "default argument must be of type int or float"
    return _save_cast(os.environ, key, __builtin__.float, __builtin__.float(default))


def string(key, default=""):
    assert isinstance(default, six.string_types), "default argument must be of type string"
    res = os.environ.get(key, default)
    invalid_chars = "\"' "
    res = res.strip(invalid_chars)
    return res


def bool(key, default=False):
    assert isinstance(default, __builtin__.bool), "default argument must be of type bool"
    val = get(key, None)
    if val is None:
        return default
    return val.upper().strip('\'" .,;-_') in ["T", "TRUE", "1", "Y", "YES", "J", "JA"]


def get(key, default=""):
    return os.environ.get(key, default)


def log_level():
    log_levels = {
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
        "WARN": logging.WARNING,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "ERR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
        "CRIT": logging.CRITICAL,
        "FATAL": logging.FATAL
    }
    return log_levels.get(os.environ.get("LOG_LEVEL", "INFO").upper(), logging.INFO)


def from_yaml(path, cls):
    if os.path.isdir(path):
        logging.error("config file is a directory, umad?")
        return cls()
    if not os.path.isfile(path):
        return cls()
    with open(path) as f:
        y = f.read().strip()
    d = related.from_yaml(y)
    return related.to_model(cls, d)


def version():
    version = string("VERSION", "")
    if not version:
        if os.path.isfile("VERSION"):
            with open("VERSION", "r") as f:
                version = f.read()
    if not version:
        version = "unknown"
    return version



import json

from redis import Redis


class RedisKV(object):
    def __init__(self, config):
        self.db = Redis(host=config.REDIS_HOST, port=config.REDIS_PORT, db=config.REDIS_DB,
                        socket_timeout=5, socket_connect_timeout=5, decode_responses=True)

    def get(self, k):
        res = self.db.hget("kv", k)
        if not res:
            return dict()
        try:
            return json.loads(res)
        except:
            return dict()

    def set(self, k, v):
        v = json.dumps(v)
        self.db.hset("kv", k, v)

    def ls(self):
        return self.db.hkeys("kv")

    def delete(self, k):
        self.db.hdel("kv", k)



from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))


def line(f):
    return f.readline().replace("\n", "").strip()


with open(path.join(here, "pip", "VERSION")) as f:
    VERSION = line(f)

with open(path.join(here, "pip", "BUILD_INFO")) as f:
    GIT_BRANCH = line(f)
    GIT_COMMIT = line(f)
    GIT_COMMIT_SHORT = line(f)

setup(
    name='arivo.om',
    version='{}'.format(VERSION),
    license="GPL-3",
    description='Util functions for om docker containers',
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    python_requires='>=2.6, !=3.0.*, !=3.1.*, !=3.2.*, <4',
    long_description="VERSION={}\n"
                     "GIT_BRANCH={}\n"
                     "GIT_COMMIT={}\n"
                     "GIT_COMMIT_SHORT={}".format(VERSION, GIT_BRANCH, GIT_COMMIT, GIT_COMMIT_SHORT),

    packages=find_packages(exclude=["om.utils.test", "om.message.test"]),
    install_requires=['six', 'related', 'enum34']
)



# coding=utf-8
from __future__ import unicode_literals

import calendar
import datetime
import logging
import os
import sys
import threading
import time
import traceback
from collections import defaultdict

import six
import zmq
from dateutil.parser import parse
from om.message import Message, Topic
from om.utils.charset import CharsetConverter


class Category(object):
    BOOKED_DIGIMON = "booked-digimon"
    BOOKED_EMPLOYEE = "booked-employee"
    BOOKED_VISITOR = "booked-visitor"
    PERMANENT_DIGIMON = "permanent-digimon"
    PERMANENT_EMPLOYEE = "permanent-employee"
    FILLER_EMPLOYEE = "filler-employee"
    FILLER_DIGIMON = "filler-digimon"
    FILLER_VISITOR_BUTTON = "filler-visitor-button"
    FILLER_VISITOR_UNEXPECTED = "filler-visitor-unexpected"

    UNKNOWN_CATEGORY = "unknown-category"


class ExternalAuthorizationProvider(object):
    USER_PREFIX = ""

    VALID_MEDIUM_TYPES = ["*"]
    VALID_REASONS = ["unknown_medium", "authority"]
    LAST_RESULT_CACHE_ENABLED = True

    def __init__(self, config):
        assert self.USER_PREFIX, ("please set USER_PREFIX")
        assert not all(x in self.USER_PREFIX for x in ("-", "_")), ("user prefix must not contain -, _")

        self.config = config
        self.context = zmq.Context()
        self.sub = self.context.socket(zmq.SUB)
        self.pub = self.context.socket(zmq.PUB)
        self.sub.connect("tcp://{}:{}".format(config.BROKER_HOST, config.BROKER_PUB_PORT))
        self.pub.connect("tcp://{}:{}".format(config.BROKER_HOST, config.BROKER_SUB_PORT))
        self.sub.subscribe(Topic.COUNT)
        self.sub.subscribe(Topic.ACCESS_REJECT)

        self.cache = defaultdict(tuple)

    def user_id_for_identifier(self, category, user_identifier):
        return "{}-{}_{}".format(self.USER_PREFIX, category, user_identifier)

    def process_reject_message(self, message):
        """
        decide if we want to process the specific reject message
        this default implementation only checks on authority and unknown_medium reasons and
        only if the user is on the loop if a loop is available

        :param message: the full access_reject message
        :return: true if we want to proceed, false otherwise
        """

        # if we have a presence loop we only check if the user is actually on the loop
        if message.get("has_presence_loop") and not message.get("on_presence_loop"):
            return False

        # only check for spefific decisions
        reason = message.get("reason")
        if reason not in self.VALID_REASONS:
            return False

        # only check for specific medium types, or all if it is set to wildcard (*)
        if "*" not in self.VALID_MEDIUM_TYPES and message.get("medium_type") not in self.VALID_MEDIUM_TYPES:
            return False

        return True

    def _convert_to_timestamp(self, dt):
        if dt is None:
            return None

        if isinstance(dt, six.string_types):
            dt = parse(dt)

        if isinstance(dt, datetime.datetime):
            dt = calendar.timegm(dt.timetuple())

        if isinstance(dt, float):
            dt = int(dt)

        if not isinstance(dt, six.integer_types):
            raise ValueError("cannot convert '{}' to timestamp".format(dt))

        return dt

    def unknown_access(self, user_id, user_category, category, zone=None):
        assert category in [Category.FILLER_VISITOR_UNEXPECTED, Category.FILLER_VISITOR_BUTTON]
        start = time.time()
        end = None
        return self.access(user_id, user_category, start, category, end, zone=zone)

    def access(self, user_id, user_category, start, category, end=None, recurrence=None, zone=None):
        result = {
            "category": category,
            "start": self._convert_to_timestamp(start),
            "end": self._convert_to_timestamp(end),
            "user": self.user_id_for_identifier(user_category, user_id),
            "recurrence": None,
            "duration": 0,
            "occupant_check": False,
        }

        if zone:
            result["zone"] = zone

        return result

    def report_conncetion_status(self, online, delay):
        pass

    def has_access(self, type, id, gate, direction, message):
        raise NotImplementedError()

    def check_in(self, user, medium_id, medium_type, timestamp, gate, direction, message):
        raise NotImplementedError()

    def check_out(self, user, medium_id, medium_type, timestamp, gate, direction, message):
        raise NotImplementedError()

    def preprocess_medium(self, id):
        if not hasattr(self, "charset"):
            self.charset = CharsetConverter.from_yaml(self.config.YAML.charset)
        return self.charset.clean(id)

    def is_in_cache(self, id, type, gate):
        return self.cache[gate][:2] == (id, type)

    def get_cached_result(self, gate):
        return self.cache[gate][2]

    def set_cache(self, id, type, gate, result):
        self.cache[gate] = (id, type, result)

    def clear_cache(self, gate):
        try:
            del self.cache[gate]
        except KeyError:
            pass

    def _handle_access_reject(self, message):
        if not self.process_reject_message(message):
            return

        medium_id, type = self.preprocess_medium(message.get("id")), message.get("medium_type")
        gateway = message.get("gateway", {})
        gate, direction = gateway.get("gate"), gateway.get("direction")
        result = False
        start = time.time()

        if self.is_in_cache(medium_id, type, gate):
            logging.info("result for {} is in cache".format((medium_id, type, gate)))
            result = self.get_cached_result(gate)
        else:
            try:
                result = self.has_access(type, medium_id, gate, direction, message)
            except Exception:
                traceback.print_exc()
                self.report_conncetion_status(False, time.time() - start)
                self.clear_cache(gate)
            else:
                self.report_conncetion_status(True, time.time() - start)
                self.set_cache(medium_id, type, gate, result)

        if result is False:
            logging.info("id {} not allowed at {}".format(medium_id, gate))
        else:
            logging.info("id {} allowed at {}".format(medium_id, gate))
            result["id"] = medium_id
            message = Message(
                name=self.config.NAME,
                type="auth",
                id=medium_id,
                medium_type=type,
                gateway=message.get("gateway"),
                access=result
            )
            message.to_socket(self.pub, Topic.BACKEND)
            logging.debug("sending: {}".format(message))

    def _handle_count_in(self, message):
        if not message.get("user", "").startswith(self.USER_PREFIX):
            return

        gateway = message.get("gateway", {})
        start = time.time()
        try:
            res = self.check_in(
                user=message.get("user"),
                medium_id=message.get("id"),
                medium_type=message.get("medium_type"),
                timestamp=message.get("timestamp"),
                gate=gateway.get("gate"),
                direction=gateway.get("direction"),
                message=message
            )
        except Exception:
            traceback.print_exc()
            self.report_conncetion_status(False, time.time() - start)
            return False
        else:
            self.report_conncetion_status(True, time.time() - start)
            return res

    def _handle_count_out(self, message):
        if not message.get("user", "").startswith(self.USER_PREFIX):
            return

        if message.get("gateway", {}).get("gate") == "automatically_deleted":
            return

        gateway = message.get("gateway", {})
        start = time.time()
        try:
            res = self.check_out(
                user=message.get("user"),
                medium_id=message.get("id"),
                medium_type=message.get("medium_type"),
                timestamp=message.get("timestamp"),
                gate=gateway.get("gate"),
                direction=gateway.get("direction"),
                message=message
            )
        except Exception:
            traceback.print_exc()
            self.report_conncetion_status(False, time.time() - start)
            return False
        else:
            self.report_conncetion_status(True, time.time() - start)
            return res

    def run_as_thread(self):
        thread = threading.Thread(target=self._thread_wrapper)
        thread.setDaemon(True)
        thread.start()
        return thread

    def _thread_wrapper(self):
        try:
            self.run()
        except Exception as e:
            logging.error("exception in auth provider thread: {}".format(str(e)))
            traceback.print_exc(file=sys.stderr)
            sys.stderr.flush()
            time.sleep(0.1)
            os._exit(1)

    def run(self):
        while True:
            topic, message = Message.from_socket(self.sub)
            if topic == Topic.ACCESS_REJECT:
                self._handle_access_reject(message)
            elif topic == Topic.COUNT:
                direction = message.get("gateway", {}).get("direction")
                if direction == "in":
                    self._handle_count_in(message)
                elif direction == "out":
                    self._handle_count_out(message)
                else:
                    logging.error("unknown direction {} for message {}".format(direction, message))



# coding=utf-8
from __future__ import unicode_literals

import json
import time


class Topic(object):
    MEDIUM = b"medium"  # alpr, nfc, command
    ACCESS_ACCEPT = b"access_accept"  # open
    ACCESS_REJECT = b"access_reject"  # e.g. if number plate is not found in database
    SENSOR = b"sensor"  # induction loop e.g for presence or counting
    COUNT = b"count"  # topic for the actual count how many vehicles are e.g. in the parking lot
    HEALTH = b"healthz"  # healthz checks
    IO = b"io"  # io events topic
    PAYMENT = b"payment"  # payment topic
    NOTICE = b"notice"  # topic for text/audio not
    GROUPING = b"grouping"  # topic for alpr grouping mechanism
    OPEN = b"ioopen"  # topic for permanently opening a gate
    PORTER = b"porter"  # topic for porter commands
    ALPR_RESULT = b"alpr_result"  # topic for sending alpr result to tracking
    TRACKING_COUNT = b"tracking_count"  # topic for tracking based counting
    TRACKING_DEBUG = b"tracking_debug"  # topic to trigger image save when new track
    BACKEND = b"backend"  # topic used for external authorization providers
    WEBAPP = b"webapp"  # topic used for webapp access
    TO_DISPLAY = b"to_display"  # show specific message on display


class Message(dict):
    _MANDATORY_FIELDS = ["name", "type"]

    @classmethod
    def parse(cls, msg):
        try:
            parsed = json.loads(msg.encode())
        except Exception as e:
            raise ValueError("Could not parse message")
        if not all(k in parsed for k in cls._MANDATORY_FIELDS):
            raise KeyError("Mandatory keys not found in message")
        return parsed

    @classmethod
    def serialize(cls, msg):
        if not isinstance(msg, dict):
            raise TypeError("Message is not of type dict")
        if not all(k in msg for k in cls._MANDATORY_FIELDS):
            raise KeyError("Mandatory keys not found in message")
        try:
            msg.update(timestamp=time.time())
            serialized = json.dumps(msg)
        except (TypeError, ValueError):
            raise ValueError("Could not serialize message")
        return serialized

    def __init__(self, name=None, type=None, *args, **kwargs):
        kwargs["type"] = type
        if "__prefix" in kwargs:
            kwargs["name"] = name
            del kwargs["__prefix"]
        else:
            kwargs["name"] = "{}-{}".format(type, name)
        super(Message, self).__init__(*args, **kwargs)

    def to_json(self):
        return Message.serialize(self)

    def from_json(self, serialized):
        self.clear()
        self.update(Message.parse(serialized))

    def to_socket(self, sock, topic):
        sock.send_multipart((topic, self.to_json()))

    @classmethod
    def from_socket(cls, sock):
        topic, payload = sock.recv_multipart()
        # beware: __prefix=False will be supplied as kwargs parameter in the __init__ function
        message = cls(__prefix=False, **cls.parse(payload))
        return topic, message

    def set_gateway(self, gate, direction=None):
        gateway = dict(gate=(str(gate) or "unknown").lower(),
                       direction=(str(direction) or "").lower())
        self.update(gateway=gateway)

    def is_gateway(self):
        return "gateway" in self

    def __str__(self):
        return self.to_json()

    def __repr__(self):
        return str(self)



import logging
import sys

from .healthz import Healthz, HealthzStatus

if sys.version_info < (3, 0):
    from .config import log_level as __log_level


    def setup_logging():
        log_level = __log_level()
        logging.basicConfig(
            level=log_level,
            format="%(asctime)s:%(levelname)s:%(name)s -> %(message)s [%(filename)s:%(lineno)s]"
        )
else:
    print("config env vars are not available in python3, use environs")

__all__ = ['Healthz', 'setup_logging']
